package example;

import java.io.File;
import java.text.SimpleDateFormat;
import java.util.Calendar;
import java.util.Locale;
import java.util.TimeZone;
import javax.swing.UIManager;
import uk.co.agena.minerva.model.MarginalDataItem;
import uk.co.agena.minerva.model.MarginalDataItemList;
import uk.co.agena.minerva.model.MarginalDataStore;
import uk.co.agena.minerva.model.Model;
import uk.co.agena.minerva.model.corebn.CoreBN;
import uk.co.agena.minerva.model.corebn.CoreBNException;
import uk.co.agena.minerva.model.corebn.CoreBNInconsistentEvidenceException;
import uk.co.agena.minerva.model.extendedbn.ContinuousEN;
import uk.co.agena.minerva.model.extendedbn.ExtendedBN;
import uk.co.agena.minerva.model.extendedbn.ExtendedNode;
import uk.co.agena.minerva.model.extendedbn.ExtendedStateException;
import uk.co.agena.minerva.model.extendedbn.ExtendedStateNumberingException;
import uk.co.agena.minerva.model.scenario.Scenario;
import uk.co.agena.minerva.util.io.FileHandlingException;
import uk.co.agena.minerva.util.model.DataSet;
import uk.co.agena.minervaapps.basicminerva.MinervaMainFrame;

/**
 * This example assumes the use of a model file named 'Adaptive Learning Segment.cmp'.
 * 
 * This model had 3 nodes:
 * - Failure_Rate_Input which is a Gamma(0.001, 1000) as the initial distibution.
 * - Failure_Rate_T which is an Identity(Failure_Rate_Input)
 * - TBF_0 which is exponential(Failure_Rate_T)
 * 
 * You can find these files in resources\api\api example models
 * 
 * Need to ensure it has 1 empty scenario.
 * 
 * The aim of this example is to show how to create an adaptive learning process via
 * the AgenaRisk API.  The process to acheive this is as follows:
 * 
 * 1) Enter an observation into the TBF node.
 * 2) Propagate the model.
 * 3) Get the posterior distribution from the Failure_Rate_T node.
 * 4) Reset the model.
 * 5) Set the prior on the Failure_Rate_Input node to be the distribution from (3).
 * 6) Go back to (1).
 * Do this as many times as necessary to use all the observations for the TBF node.
 * 
 *
 */
public class AdaptiveLearning {
  
  final public static String MODEL_FILENAME = "Adaptive Learning Segment.cmp";
  final public static String ADAPTIVE_EXTENDEDBN_NAME = "Failure Rate Adaptive";
  final public static String ADAPTIVE_INPUT_EXTENDEDNODE_UNIQUEID = "Failure_Rate_Input";
  final public static String ADAPTIVE_OUTPUT_EXTENDEDNODE_UNIQUEID = "Failure_Rate_T";
  final public static String EVIDENCE_EXTENDEDNODE_UNIQUEID = "TBF_0";
  final public static double[] evidence = {100.0, 150.0, 200.0, 200.0, 200.0, 250.0, 300.0, 350.0, 400.0};
  
  /**
   * This is the model that will learn.
   */
  Model model;
  
  /**
   * This is the extendedbn that has the learning nodes.
   */
  ExtendedBN extendedbn;
  
  /**
   * This is the node which will have the prior distibution set from the posterior.
   */
  ExtendedNode adaptiveInputNode;
  
  /**
   * This is the node which will have the posterior distibution extracted from it.
   */
  ExtendedNode adaptiveOutputNode;
  
  /**
   * This is the node which will have the evidence repeatedly entered on it.
   */
  ExtendedNode evidenceNode;
  


  public static void main(String[] args) {
    try {
      AdaptiveLearning al = new AdaptiveLearning();
      al.learn();
      
    }catch(Exception e) {
      e.printStackTrace();
    }
    
  }
  
  AdaptiveLearning() throws FileHandlingException{
    //First load the model
    this.model = Model.load(MODEL_FILENAME);
    
    //Set up access to all the variables which we will require.
    this.extendedbn = this.model.getExtendedBNWithName(ADAPTIVE_EXTENDEDBN_NAME);
    this.adaptiveInputNode = this.extendedbn.getExtendedNodeWithUniqueIdentifier(ADAPTIVE_INPUT_EXTENDEDNODE_UNIQUEID);
    this.adaptiveOutputNode = this.extendedbn.getExtendedNodeWithUniqueIdentifier(ADAPTIVE_OUTPUT_EXTENDEDNODE_UNIQUEID);
    this.evidenceNode = this.extendedbn.getExtendedNodeWithUniqueIdentifier(EVIDENCE_EXTENDEDNODE_UNIQUEID);
  }
  
  private void learn() 
      throws Exception {
    
    //Get the scenario
    Scenario scenario = (Scenario) this.model.getScenarioList().getScenarios().get(0);
    
    //Set a place holder for the prior for the adaptive extendednode. (First iteration doesn't require one).
    DataSet adaptiveInputNodePrior = null;
    
    // Go through each observation entry
    for(int i=0; i<evidence.length; i++) {
      //Make sure there are no observations
      scenario.clearAllObservations();
      //Enure there is no evidence propagated into the model
      extendedbn.retractAllEvidence(false);
      
      if(adaptiveInputNodePrior!=null) {
        //Set the prior on the adaptive extended node
        setPrior(adaptiveInputNode, adaptiveInputNodePrior);
      }
      
      double evidenceValue = evidence[i];
      //Create a new observation for the evidence
      scenario.addRealObservation(extendedbn.getId(),evidenceNode.getId(), evidenceValue);
      
      
      //Propagate the model
      model.calculate();
      
      //Get the posterior for the adaptive extendedNode
      DataSet posterior = getPosterior(adaptiveOutputNode);
      adaptiveInputNodePrior = (DataSet) posterior.clone();
      
    }//end for i/evidence
    
    showGUI();
  }

  
  /**
   * We now set the prior on the argument node.  Ensuring that:
   * - It is not a Dynamic Discretisable Node (if its a continousEN node).
   * - It has its NPT set to Manual, i.e. not generated from a function.
   * 
   * @param enode
   * @param prior
   * @throws ExtendedStateNumberingException
   * @throws ExtendedStateException
   * @throws CoreBNException
   * @throws CoreBNInconsistentEvidenceException
   */
  private void setPrior(ExtendedNode enode, DataSet prior) throws ExtendedStateNumberingException, ExtendedStateException, CoreBNException, CoreBNInconsistentEvidenceException {
    //If continous then turn of DD
    if(enode instanceof ContinuousEN) {
      ContinuousEN contEN = (ContinuousEN) enode;
      contEN.setDynamicallyDiscretisable(false);
    }
    //Set it as a manual function
    enode.setFunctionMode(ExtendedNode.EDITABLE_NPT);
    
    //Now set the states and priors on this node
    enode.createExtendedStates(prior);
    CoreBN connBn = this.extendedbn.getConnBN();
    String adaptiveNode_CoreNodeId = enode.getConnNodeId();
    connBn.touchTable(adaptiveNode_CoreNodeId, prior.getAsDoubles(), false);
  }

  /**
   * Returns the posterior distibution (dataset) for the argument node.
   * @param enode
   * @return
   */
  private DataSet getPosterior(ExtendedNode enode) {
    //A marginal data store holds the calculated posterior results for all the nodes.
    MarginalDataStore mds = model.getMarginalDataStore();
    //A 'marginal data item List' is a list of calculated posteriors for this node.
    MarginalDataItemList mdilist = mds.getMarginalDataItemListForNode(extendedbn, enode);
    //We only require the first item, as this is the results for the first scenario (we only have one).
    MarginalDataItem mdi = mdilist.getMarginalDataItemAtIndex(0);
    
    //The Dataset is a store of the posterior values for each state of the node.
    return mdi.getDataset();
  }
  
  
  
  private void showGUI() throws Exception {
    if (this.model == null)
        throw new Exception("Cannot open non-existant model");
    
    Calendar cal = Calendar.getInstance(TimeZone.getDefault());
    String DATE_FORMAT = "yyyyMMddHHmmss";
    SimpleDateFormat sdf = new SimpleDateFormat(DATE_FORMAT);
    String time = sdf.format(cal.getTime());
    String filename = time+MODEL_FILENAME;
    try {
        model.save(filename);
    } catch (FileHandlingException e) {
        throw new Exception("Could not save model to:" + filename, e);
    }
    
    try {
        Locale.setDefault(Locale.ENGLISH);
        UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
        MinervaMainFrame mf = new MinervaMainFrame("", false, true);
        mf.setVisible(true);
        mf.openModel(filename,false,false,false);
        File f = new File(filename);
        f.delete();
    } catch (Exception e) {
        throw new Exception("There was a problem opening the application. : " + e.getMessage(), e);
    }
  }

}
